/* file r_job.c */
/*
 *      dio87d (Diogene87) - a powerful job scheduling system.
 *	Copyright (C) 2003,2004,2005,2006,2007 - Dominique Parisot
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 *
 */

#include <config.h>
#include "gettext.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>
#include <signal.h>

#if DG87_SOLARIS == 1
	#include <sys/filio.h>
#endif

#include "r_caller.h"
#include "r_job.h"


extern char **environ;          /* environment array */


/****************************************************************************/
BOOL
dg87_r_job_change_state (dg87_r_job * job, dg87_r_job_state state)
{
  assert (job != NULL);
  if (job == NULL)
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return FALSE;
    }
  dg87_daemon_log_printf (dg87_daemon_log_debug,
                          "set job state %s %s %s :%s",
                          job->day_name, job->task_name, job->job_name,
                          dg87_r_job_state_to_ascii (state));
  job->state = state;
  job->transmit_state = TRUE;
  DG87_SAVE_NEEDED;
  return TRUE;
}

/****************************************************************************/
BOOL
dg87_r_job_transmit_state (dg87_context * context, dg87_r_job * job)
{
  assert (job != NULL);
  if (job == NULL)
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return FALSE;
    }
  if (job->transmit_state == TRUE)
    {
      dg87_daemon_log_printf (dg87_daemon_log_debug,
                              "send job state %s :%s", job->job_name,
                              dg87_r_job_state_to_ascii (job->state));
      if (dg87_r_caller_transmit_state (context, job) == TRUE)
        {
          job->transmit_state = FALSE;
          if (job->state == dg87_r_job_state_exit ||
              job->state == dg87_r_job_state_deleted)
            job->state = dg87_r_job_state_old;
          DG87_SAVE_NEEDED;
        }
    }
  return TRUE;
}

/****************************************************************************/
void
dg87_r_job_array_cleanup (char **argv)
{
  char **arg;
  assert (argv != NULL);
  if (argv == NULL)
    return;
  arg = argv;
  while (*arg != NULL)
    {
      dg87_free (*arg, DG87_CHAR);
      arg++;
    }
  dg87_free (argv, DG87_CHAR);
}

/****************************************************************************/
char **
dg87_r_job_split_argument (char *path, char *argument)
{
  char **argv = NULL;
  int argc = 0;
  int new_size;
  char **new_buffer;
  char *beg;
  char *end;
  int arg_size;

  assert (path != NULL);
  if (path == NULL)
    return NULL;

  /* first element in array is job path */
  argv = dg87_alloc (sizeof (char *) * 2, DG87_CHAR);
  if (argv == NULL)
    return NULL;
  argv[0] = dg87_alloc (sizeof (char) * strlen (path) + 1, DG87_CHAR);
  if (argv[0] == NULL)
    {
      dg87_free (argv, DG87_CHAR);
      return NULL;
    }
  strcpy (argv[0], path);
  argv[1] = NULL;               /* terminator */
  argc = 1;

  if (argument == NULL)
    return argv;

  end = argument;
  while (1)
    {

      /* skip separators */
      beg = end;
      while (*beg == ' ' && *beg != '\0')
        beg++;
      if (*beg == '\0')
        break;

      /* if character is quote */
      end = beg + 1;
      if (*beg == '\"')
        {
          /* skip until next quote */
          while (*end != '\"' && *end != '\0')
            end++;
          if (*end == '\0')
            {
              dg87_r_job_array_cleanup (argv);
              dg87_err_set_error (DG87_ERR_ARGUMENT_SYNTAX_ERROR);
              return NULL;
            }
          /* remove first quote */
          beg++;
          end++;
          arg_size = end - beg - 1;
        }
      else
        {
          /* else skip word */
          while (*end != ' ' && *end != '\0')
            end++;
          arg_size = end - beg;
        }

      /* increase argument array size */
      new_size = argc + 1;
      new_buffer = realloc (argv, sizeof (char *) * (new_size + 1));
      if (new_buffer == NULL)
        {
          dg87_r_job_array_cleanup (argv);
          dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
          return NULL;
        }
      argv = new_buffer;
      argc = new_size;
      argv[argc] = NULL;        /* terminator */

      /* add argument in array */
      argv[argc - 1] =
        dg87_alloc (sizeof (char) * (arg_size + 1), DG87_CHAR);
      if (argv[argc - 1] == NULL)
        {
          dg87_r_job_array_cleanup (argv);
          dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
          return NULL;
        }
      memset (argv[argc - 1], 0, sizeof (char) * (arg_size + 1));
      dg87_strnzcpy (argv[argc - 1], beg, arg_size);
    }

  return argv;
}

/****************************************************************************/
BOOL
dg87_r_job_run (dg87_r_job * job)
{
  char **argv = NULL;
  int filedes[2];
  dg87_strbuf *msg;
  char *hostname;

  assert (job != NULL);
  if (job == NULL)
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return FALSE;
    }

  /* create a pipe for communication between dio87d and the job. */
  /* fildes[0] is for reading, fildes[1] is for writing*/
  if (pipe (filedes) == -1)
    {
  		dg87_r_caller_send_log_entry (job->context, job, 
					dg87_log_day_entry_type_info, gettext("Can't create a pipe !"));
  		dg87_r_job_change_state (job, dg87_r_job_state_deleted);
      dg87_err_set_error (DG87_ERR_CAN_NOT_FORK);
      return FALSE;
    }


  /* send log entries */
  if ((msg = dg87_strbuf_new ()) == NULL)
    return FALSE;

  /* path entry */
  dg87_strbuf_printf (msg, gettext("Path:%s"), job->path);
  dg87_r_caller_send_log_entry (job->context, job, dg87_log_day_entry_type_info,
                                msg->buffer);
  /* argument entry */
  dg87_strbuf_cpy (msg, "");
  dg87_strbuf_printf (msg, gettext("Arguments:%s"), job->argument);
  dg87_r_caller_send_log_entry (job->context, job, dg87_log_day_entry_type_info,
                                msg->buffer);
  /* hostname entry */
  hostname = dg87_context_get_hostname (job->context, FALSE);
  if (hostname == NULL)
    {
      dg87_strbuf_delete (msg);
      return FALSE;
    }
  dg87_strbuf_cpy (msg, "");
  dg87_strbuf_printf (msg, gettext("Hostname:%s"), hostname);
  dg87_free (hostname, DG87_CHAR);
  dg87_r_caller_send_log_entry (job->context, job, dg87_log_day_entry_type_info,
                                msg->buffer);

  job->pid = fork ();
  if (job->pid == -1)
    {
      /* can't fork */
  		dg87_r_caller_send_log_entry (job->context, job, 
					dg87_log_day_entry_type_info, 
					gettext("Can't create a process (fork error)!"));
  		dg87_r_job_change_state (job, dg87_r_job_state_deleted);
      dg87_daemon_log_printf (dg87_daemon_log_error,
                          gettext("fork error for the job %s %s %s"),
                          job->day_name, job->task_name, job->job_name);
      dg87_err_set_error (DG87_ERR_CAN_NOT_FORK);
      return FALSE;
    }


  if (job->pid == 0)
    {
      /***** this code is executed by the child *****/
  		/* PID entry */
			job->pid = getpid();
  		dg87_strbuf_cpy (msg, "");
  		dg87_strbuf_printf (msg, gettext("PID:%i"), job->pid);
  		dg87_r_caller_send_log_entry (job->context, job, 
						dg87_log_day_entry_type_info, msg->buffer);

      /* define environment variables */
      setenv ("DG87_DAY", job->day_name, 1);
      setenv ("DG87_TASK", job->task_name, 1);
      setenv ("DG87_JOB", job->job_name, 1);

      /* redirect stdin and stdout to the pipe */
      if (close (1) == -1 || close(2) == -1)
      {
          dg87_r_job_array_cleanup (argv);
          dg87_free (hostname, DG87_CHAR);
          return FALSE;
      }
      /* the child process closes the unused read end of the pipe,
       * because it only sends outputs to dio87d */ 
      close(filedes[0]);

      /* connect stdout to pipe */
      if (dup2 (filedes[1], 1) == -1)
      {
          dg87_r_job_array_cleanup (argv);
          dg87_free (hostname, DG87_CHAR);
          return FALSE;
      }
      /* connect stderr to pipe */
      if (dup2 (filedes[1], 2) == -1)     
      {
          dg87_r_job_array_cleanup (argv);
          dg87_free (hostname, DG87_CHAR);
          return FALSE;
      }

      argv = dg87_r_job_split_argument (job->path, job->argument);
      if (argv == NULL)
        {
          dg87_r_job_array_cleanup (argv);
          dg87_free (hostname, DG87_CHAR);
          return FALSE;
        }

      /* execute job */
      execve (job->path, argv, environ);
      dg87_daemon_log_printf (dg87_daemon_log_warning,
                          gettext("execve error for the job %s %s %s : %s"),
                          job->day_name, job->task_name, job->job_name,
													strerror(errno));
      dg87_r_job_array_cleanup (argv);

      /* execve return only if an error occures */
      /* kill the child process */
      close (1);
      close (2);

			/* add an entry in the job log */
  		dg87_strbuf_cpy (msg, "");
  		dg87_strbuf_printf (msg, 
				gettext("The job can not be executed ! %s"), strerror(errno));
  		dg87_r_caller_send_log_entry (job->context, job, 
					dg87_log_day_entry_type_info, msg->buffer);
  		dg87_r_job_change_state (job, dg87_r_job_state_deleted);

      /* stop child process execution */
      exit(DELETED_EXIT_CODE);
    }

	dg87_strbuf_delete (msg);
  /* the dio87d daemon closes the unused write end of the pipe,
  * because it only reads outputs of the child process*/ 
  close(filedes[1]);
  job->fd = filedes[0];

  dg87_r_job_change_state (job, dg87_r_job_state_running);

  return TRUE;
}

/****************************************************************************/
BOOL
dg87_r_job_signal (dg87_r_job * job, int signal)
{
	char *msg;
	/* send a signal to the job */
	if (kill(job->pid, signal) == 0)
		msg = gettext("Send a signal to the job");
	else
		msg = gettext("Can't send a signal to the job");
	/* add a entry in the job log */
	dg87_r_caller_send_log_entry (job->context, job, 
					dg87_log_day_entry_type_info, msg);
	return TRUE;
}

/****************************************************************************/
BOOL
dg87_r_job_catch_output (dg87_r_job * job, dg87_context * context)
{
	const int ntoread=256;
  int nread;
  char *buffer;
  char *beg, *end;
  assert (job != NULL && context != NULL);
  if (job == NULL)
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return FALSE;
    }
  if (job->fd == -1)
    return FALSE;

  /* allocate buffer */
  if ((buffer = dg87_alloc (ntoread + 1, DG87_CHAR)) == NULL)
    {
      dg87_err_set_error (DG87_ERR_OUT_OF_MEMORY);
      return FALSE;
    }
  /* read jobs output */
  nread = read (job->fd, buffer, ntoread);
  buffer[nread] = '\0';
  dg87_strbuf_ncat (job->output, buffer, nread);
  dg87_free (buffer, DG87_CHAR);

  /* send a message for each line */
  beg = job->output->buffer;
  while (1)
    {
      if (*beg == '\0')
        break;
      if ((end = strchr (beg, '\n')) == NULL)
        break;
      *end = '\0';
      dg87_r_caller_send_log_entry (context, job,
                                    dg87_log_day_entry_type_output, beg);
      beg = end + 1;
    }
  /* remove begin of buffer which corresponds to lines which are
   * sended */
  if (buffer != beg)
    {
      buffer = dg87_strbuf_get_buffer (job->output);
      dg87_strbuf_cat (job->output, beg);
      dg87_free (buffer, DG87_CHAR);
    }
  return TRUE;
}

/****************************************************************************/
dg87_r_job *
dg87_r_job_new (char *day_name, char *task_name, char *job_name,
                char *path, char *argument, char *hostname, int ip_port,
								dg87_context* context)
{
  dg87_r_job *job;
  assert (task_name != NULL && job_name != NULL && day_name != NULL
          && path != NULL && argument != NULL && hostname != NULL
					&& context != NULL);
  if (task_name == NULL || job_name == NULL || day_name == NULL
      || path == NULL || argument == NULL || hostname == NULL 
			|| context == NULL )
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return NULL;
    }
  /* Path length can not be zero */
  if (strlen (path) == 0)
    {
      dg87_err_set_error (DG87_ERR_INCORRECT_VALUE);
      return NULL;
    }
  /* check job name, task name and day name */
  if (dg87_check_name (job_name, DG87_JOB_NAME_LENGTH) == FALSE)
    return NULL;
  if (dg87_check_name (task_name, DG87_TASK_NAME_LENGTH) == FALSE)
    return NULL;
  if (dg87_check_name (day_name, DG87_DAY_NAME_LENGTH) == FALSE)
    return NULL;
  if (dg87_check_valid_characters (path) == FALSE)
    return NULL;
  if (dg87_check_valid_characters (argument) == FALSE)
    return NULL;
  if (dg87_check_valid_characters (hostname) == FALSE)
    return NULL;
  /* allocate memory */
  if ((job = dg87_alloc (sizeof (dg87_r_job), DG87_JOB)) == NULL)
    goto alloc_error;
  if ((job->path = dg87_strdup (path, DG87_CHAR)) == NULL)
    goto alloc_error;
  if ((job->argument = dg87_strdup (argument, DG87_CHAR)) == NULL)
    goto alloc_error;
  if ((job->output = dg87_strbuf_new ()) == NULL)
    goto alloc_error;
  if ((job->hostname = dg87_strdup (hostname, DG87_CHAR)) == NULL)
    goto alloc_error;
  /* initialize structure */
  dg87_strnzcpy (job->job_name, job_name, DG87_JOB_NAME_LENGTH);
  dg87_strnzcpy (job->task_name, task_name, DG87_TASK_NAME_LENGTH);
  dg87_strnzcpy (job->day_name, day_name, DG87_DAY_NAME_LENGTH);
  job->ip_port = ip_port;
	job->context = context;
  job->state = dg87_r_job_state_uninitialize;
  job->pid = -1;
  job->status = -1;
  job->fd = -1;
  DG87_SAVE_NEEDED;
  return job;

alloc_error:
  if (job != NULL)
    dg87_r_job_delete (job);
  dg87_err_set_error (DG87_ERR_OUT_OF_MEMORY);
  return NULL;
}

/****************************************************************************/
BOOL
dg87_r_job_delete (dg87_r_job * job)
{
  assert (job != NULL);
  if (job == NULL)
    {
      dg87_err_set_error (DG87_ERR_NULL_PARAMETER);
      return FALSE;
    }
  if (job->output != NULL)
    dg87_strbuf_delete (job->output);
  if (job->path != NULL)
    dg87_free (job->path, DG87_CHAR);
  if (job->argument != NULL)
    dg87_free (job->argument, DG87_CHAR);
  if (job->hostname != NULL)
    dg87_free (job->hostname, DG87_CHAR);
  if (job->fd != -1)
    close (job->fd);
  dg87_free (job, DG87_JOB);
  DG87_SAVE_NEEDED;
  return TRUE;
}

/****************************************************************************/
BOOL
dg87_r_job_load (dg87_r_job * job, dg87_storage * storage)
{
  dg87ubyte b;
  dg87uint ui;
  dg87int i;
  assert (job != NULL && storage != NULL);
  /* read 'state' property */
  if (dg87_storage_read_ubyte (storage, &b) == FALSE)
    return FALSE;
  job->state = b;
  /* read 'pid' property */
  if (dg87_storage_read_uint (storage, &ui) == FALSE)
    return FALSE;
  job->pid = ui;
  /* read 'status' property */
  if (dg87_storage_read_int (storage, &i) == FALSE)
    return FALSE;
  job->status = i;
  /* read 'transmit_state' property */
  if (dg87_storage_read_ubyte (storage, &b) == FALSE)
    return FALSE;
  job->transmit_state = b;
  return TRUE;
}

/****************************************************************************/
BOOL
dg87_r_job_save (dg87_r_job * job, dg87_storage * storage)
{
  assert (job != NULL && storage != NULL);
  /* 'state' property */
  if (dg87_storage_write_ubyte (storage, job->state) == FALSE)
    return FALSE;
  /* 'pid' property */
  if (dg87_storage_write_uint (storage, job->pid) == FALSE)
    return FALSE;
  /* 'status' property */
  if (dg87_storage_write_uint (storage, job->status) == FALSE)
    return FALSE;
  /* 'transmit_state' property */
  if (dg87_storage_write_ubyte (storage, job->transmit_state) == FALSE)
    return FALSE;
  return TRUE;
}

/****************************************************************************/
char *
dg87_r_job_state_to_ascii (dg87_r_job_state state)
{
  switch (state)
    {
    case dg87_r_job_state_deleted:
      return "deleted";
    case dg87_r_job_state_running:
      return "running";
    case dg87_r_job_state_exit:
      return "exit";
    case dg87_r_job_state_old:
      return "old";
    default:
      return "?";
    }
  return "?";
}

/****************************************************************************/
#ifdef DEBUG
int
dg87_r_job_check (dg87_r_job * job, int (*print) (char *))
{
  if (strlen (job->job_name) == 0)
    {
      print ("Job : job name is zero length !\n");
      return 1;
    }
  if (strlen (job->task_name) == 0)
    {
      print ("Job : task name is zero length !\n");
      return 1;
    }
  if (strlen (job->day_name) == 0)
    {
      print ("Job : day name is zero length !\n");
      return 1;
    }
  if (job->path == NULL)
    {
      print ("Job : path is NULL !\n");
      return 1;
    }
  if (strlen (job->path) == 0)
    {
      print ("Job : path is zero length !\n");
      return 1;
    }
  if (job->argument == NULL)
    {
      print ("Job : argument is NULL !\n");
      return 1;
    }

  return 0;
}
#endif /*DEBUG*/
/****************************************************************************/
/* End of file */
